@@ -15,6 +15,9 @@
  * You should have received a copy of the GNU Lesser General
  * Public License along with this library; if not, see
  * <http://www.gnu.org/licenses/>.
+ *
+ * In addition, when the library is used with OpenSSL, a special
+ * exception applies. Refer to the LICENSE_EXCEPTION file for details.
  */
 
 #include "config.h"
@@ -215,7 +218,8 @@ g_tls_connection_gnutls_init_priorities (void)
     {
       g_warning ("G_TLS_GNUTLS_PRIORITY is invalid; ignoring!");
       base_priority = DEFAULT_BASE_PRIORITY;
-      gnutls_priority_init (&priorities[FALSE][FALSE], base_priority, NULL);
+      ret = gnutls_priority_init (&priorities[FALSE][FALSE], base_priority, NULL);
+      g_warn_if_fail (ret == 0);
     }
 
   unsafe_rehandshake_priority = g_strdup_printf ("%s:%%UNSAFE_RENEGOTIATION", base_priority);
@@ -254,6 +258,8 @@ g_tls_connection_gnutls_init_priorities (void)
       fallback_priority = g_strdup_printf ("%s:%%COMPAT:!VERS-TLS-ALL:+VERS-%s",
 					   cleaned_base,
 					   gnutls_protocol_get_name (fallback_proto));
+
+      g_free (cleaned_base);
     }
   fallback_unsafe_rehandshake_priority = g_strdup_printf ("%s:%%UNSAFE_RENEGOTIATION",
 							  fallback_priority);
@@ -1274,7 +1280,7 @@ accept_peer_certificate (GTlsConnectionGnutls *gnutls,
 			 GTlsCertificate      *peer_certificate,
 			 GTlsCertificateFlags  peer_certificate_errors)
 {
-  gboolean accepted;
+  gboolean accepted = FALSE;
 
   if (G_IS_TLS_CLIENT_CONNECTION (gnutls))
     {
@@ -1283,14 +1289,9 @@ accept_peer_certificate (GTlsConnectionGnutls *gnutls,
 
       if ((peer_certificate_errors & validation_flags) == 0)
 	accepted = TRUE;
-      else
-	{
-	  accepted = g_tls_connection_emit_accept_certificate (G_TLS_CONNECTION (gnutls),
-							       peer_certificate,
-							       peer_certificate_errors);
-	}
     }
-  else
+
+  if (!accepted)
     {
       accepted = g_tls_connection_emit_accept_certificate (G_TLS_CONNECTION (gnutls),
 							   peer_certificate,
